#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
import os

from aiohttp.web import run_app

from iss_positioner import ISSPositionerService, util, LOG_FORMAT

DIR = os.path.join(os.path.dirname(__file__))
CFG = util.load_cfg(path=os.path.join(DIR, 'iss-positioner.yml'))

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)
    app = ISSPositionerService(config=CFG)
    run_app(app, port=8081)



# -*- coding: utf-8  -*-
from aiohttp import ClientSession

from .log import logger
from .redis import fill_db

__all__ = (
    'listen_tle_storage',
    'listen_tle_queue',
)


async def listen_tle_queue(queue, channel, redis, satellite, *, tle_getter=None):
    while True:
        msg = await queue.get()
        try:
            if msg.get('stored'):
                result = await fill_db(redis, satellite, tle_getter)
                added_from, added_to, added, deleted_from, deleted_to, deleted = result
                for client in channel:
                    try:
                        await client.send_json({
                            'added': {
                                'count': added,
                                'from': added_from.isoformat(' '),
                                'to': added_to.isoformat(' '),
                            },
                            'updated': {
                                'from': added_from.isoformat(' '),
                                'to': added_to.isoformat(' '),
                            },
                            'deleted': {
                                'count': deleted,
                                'from': deleted_from.isoformat(' '),
                                'to': deleted_to.isoformat(' '),
                            }
                        })
                    except Exception as exc:
                        logger.exception(exc)
        except Exception as exc:
            logger.exception(exc)
        finally:
            queue.task_done()


async def listen_tle_storage(ws_url, queue, *, loop=None):
    async with ClientSession(loop=loop) as session:
        async with session.ws_connect(ws_url) as ws:
            while True:
                msg = await ws.receive_json()
                if msg.pop('channel', None) == 'tle:nasa:update':
                    logger.info('TLE from NASA have been updated: %s', msg)
                    await queue.put(msg)



# -*- coding: utf-8 -*-
import asyncio
from collections import defaultdict
from functools import partial
from itertools import chain

from aiohttp import web, WSCloseCode
from aioredis import create_reconnecting_redis

from .calculations import AsyncSatellite
from .handlers import subscribe, index, CoordsApi, RadiusApi, LST
from .middlewares import json_response_middleware
from .tasks import listen_tle_queue, listen_tle_storage
from .util import get_tle


__all__ = (
    'ISSPositionerService',
)

async def start_background_tasks(app):
    tle_storage_conf = app['config']['tle-storage'].copy()
    redis_params = app['config']['redis'].copy()
    redis_addres = redis_params.pop('host', 'localhost'), redis_params.pop('port', 6379)

    tle_getter = partial(get_tle, url=tle_storage_conf['query_url'], loop=app.loop)

    app['satellite'] = AsyncSatellite(loop=app.loop)
    app['redis'] = await create_reconnecting_redis(redis_addres, **redis_params, loop=app.loop)

    if app['config']['on-startup']['redis']['flush']:
        await app['redis'].flushall()

    tle_storage_listener = listen_tle_storage(tle_storage_conf['listen_url'], app['queue'], loop=app.loop)
    tle_queue_listener = listen_tle_queue(app['queue'], app['channels']['coords'],
                                          app['redis'], app['satellite'],
                                          tle_getter=tle_getter)

    app['tle_storage_listener'] = app.loop.create_task(tle_storage_listener)
    app['tle_queue_listener'] = app.loop.create_task(tle_queue_listener)

    if app['config']['on-startup']['redis']['fill']:
        app['queue'].put_nowait(dict(stored=True))


async def cleanup_background_tasks(app):
    app['tle_storage_listener'].cancel()
    app['tle_queue_listener'].cancel()

    await app['tle_storage_listener']
    await app['tle_queue_listener']


async def on_shutdown(app):
    clients = chain.from_iterable(app['channels'].values())
    fs = (ws.close(code=WSCloseCode.GOING_AWAY, message='Server shutdown') for ws in clients)
    await asyncio.gather(*fs, loop=app.loop)


class ISSPositionerService(web.Application):
    def __init__(self, *, config=None, **kwargs):
        if not isinstance(config, dict):
            raise ValueError('Argument `config` must be dict()')

        super().__init__(**kwargs)

        self.middlewares.append(json_response_middleware)

        self['config'] = config
        self['channels'] = defaultdict(set)
        self['queue'] = asyncio.Queue()

        self.on_startup.append(start_background_tasks)
        self.on_cleanup.append(cleanup_background_tasks)

        self.router.add_get('/', index)
        self.router.add_get('/subscribe/{channel}', subscribe)

        self.router.add_route('*', '/lst', LST)
        self.router.add_route('*', '/coords', CoordsApi)
        self.router.add_route('*', '/radius', RadiusApi)



# -*- coding: utf-8 -*-
import asyncio
from collections import defaultdict
from datetime import datetime, timedelta
from itertools import chain

from .calculations import AsyncSatellite
from .log import tqdm
from .util import datetime_range, geo_member_to_dict

__all__ = (
    'fill_db',
    'clear_old_coords',
    'store_coords',
    'get_coords',
    'geo_radius',
)


async def clear_old_coords(redis, *, start_dt=None, end_dt=None, calculation_step=timedelta(hours=1)):
    today = datetime(*datetime.today().timetuple()[:3])
    if not end_dt:
        end_dt = today
    if not start_dt:
        start_dt = end_dt - timedelta(days=14)
    keys = (dt.isoformat(' ') for dt in datetime_range(start_dt, end_dt, calculation_step))
    deleted_count = await redis.delete(*keys)
    return deleted_count, start_dt, end_dt


async def fill_db(redis, satellite, tle_getter):
    if not callable(tle_getter):
        raise ValueError('Argument `tle_getter` must be callable object')

    satellite.tle_db = (tle['extra_info'] for tle in await tle_getter())
    return await store_coords(satellite, redis)


async def geo_radius(redis, *, dt, lon, lat, **params):
    vals = await redis.georadius(dt.isoformat(' '), lon, lat,
                                 params.get('dist', 250), unit=params.get('unit', 'km'),
                                 with_dist=True, with_coord=True, with_hash=True)
    vals.sort(key=lambda val: val.member)
    return [geo_member_to_dict(val, units=params.get('unit', 'km')) for val in vals]


async def get_coords(redis, *, start_dt, end_dt, step, calculation_step=timedelta(hours=1), loop=None):
    _start, _end = datetime(*start_dt.timetuple()[:4]), datetime(*end_dt.timetuple()[:4])

    dt_set = defaultdict(list)
    for dt in datetime_range(_start, _end, calculation_step):
        for t in datetime_range(dt, dt + calculation_step, step):
            if start_dt == end_dt and t == start_dt:
                dt_set[dt.isoformat(' ')].append(t.isoformat(' '))
                break

            if start_dt <= t < end_dt:
                dt_set[dt.isoformat(' ')].append(t.isoformat(' '))

    fs = (redis.geopos(k, *v) for k, v in dt_set.items())
    coords_set = await asyncio.gather(*fs, loop=loop)
    dt_set = zip(chain.from_iterable(dt_set.values()), chain.from_iterable(coords_set))
    return [dict(dt=dt, coords=coords._asdict()) for dt, coords in dt_set]


async def store_coords(satellite, redis, *, calculation_step=timedelta(hours=1), clear_old=True):
    if not isinstance(satellite, AsyncSatellite):
        raise ValueError('Argument `satellite` must be is instance `AsyncSatellite` class')

    dts = tuple(satellite.tle_db.keys())
    start_dt, end_dt = datetime(*datetime.today().timetuple()[:3]), dts[-1] + timedelta(days=14)
    added_count = 0
    for dt in tqdm(tuple(datetime_range(start_dt, end_dt, calculation_step)), mininterval=1):
        coords = await satellite.compute(dt, dt + calculation_step)
        added_count += await redis.geoadd(dt.isoformat(' '), *chain.from_iterable(coords))

    deleted_start, deleted_end, deleted_count = None, None, 0
    if clear_old:
        deleted_count, deleted_start, deleted_end = await clear_old_coords(redis)

    return start_dt, end_dt, added_count, deleted_start, deleted_end, deleted_count



# -*- coding: utf-8 -*-
import asyncio
from datetime import datetime, timedelta
from functools import partial
from json import JSONDecodeError

from aiohttp import web
from dateutil import parser as dt_parser

from .log import logger
from .redis import geo_radius, get_coords
from .util import datetime_range, load_html, read_lst

__all__ = (
    'index',
    'subscribe',
    'CoordsApi',
    'RadiusApi',
    'LST',
)


class CoordsApi(web.View):
    async def post(self):
        data = await get_json(self.request)
        if 'dt' in data:
            return await self.coords(data)
        return await self.coords_range(data)

    async def coords(self, data):
        start_dt = end_dt = dt_parser.parse(data.pop('dt'))
        if start_dt < datetime(*datetime.today().timetuple()[:3]):
            raise web.HTTPNotFound
        return await self.get_coords(start_dt=start_dt, end_dt=end_dt, step=1)

    async def coords_range(self, data):
        _validate_requires({'start_dt', 'end_dt'}, data, is_body=False)
        dts = dt_parser.parse(data.pop('start_dt')), dt_parser.parse(data.pop('end_dt'))
        start_dt, end_dt = min(dts), max(dts)
        if start_dt < datetime(*datetime.today().timetuple()[:3]):
            raise web.HTTPNotFound
        return await self.get_coords(start_dt=start_dt, end_dt=end_dt, step=data.get('step', 1))

    async def get_coords(self, **params):
        return await get_coords(self.request.app['redis'], loop=self.request.app.loop, **params)


class RadiusApi(web.View):
    async def post(self):
        data = await get_json(self.request)
        if 'objects' in data:
            return await self.radius_many(data)
        return await self.radius(data)

    async def radius(self, data):
        _validate_requires({'start_dt', 'end_dt', 'lat', 'lon'}, data)
        dts = dt_parser.parse(data.pop('start_dt')), dt_parser.parse(data.pop('end_dt'))
        start_dt, end_dt = min(dts), max(dts)
        if start_dt < datetime(*datetime.today().timetuple()[:3]):
            raise web.HTTPNotFound

        params = dict(start_dt=start_dt,
                      end_dt=end_dt,
                      lon=data['lon'],
                      lat=data['lat'],
                      dist=data.get('dist', 250),
                      units=data.get('units', 'km'),
                      min_duration=data.get('min_duration'))

        return await self.intersect(**params)

    async def radius_many(self, data):
        _validate_requires({'objects', 'start_dt', 'end_dt'}, data)

        objects = data['objects']
        if not isinstance(objects, list):
            raise web.HTTPBadRequest(reason='Body parameter `objects` must be `array`')
        object_keys = {'lat', 'lon'}
        if not all(object_keys.issubset(obj) for obj in objects):
            raise web.HTTPBadRequest(reason='Object inside array `objects` '
                                            'must contains required keys `{}`'.format(object_keys))

        dts = dt_parser.parse(data.pop('start_dt')), dt_parser.parse(data.pop('end_dt'))
        start_dt, end_dt = min(dts), max(dts)
        if start_dt < datetime(*datetime.today().timetuple()[:3]):
            raise web.HTTPNotFound

        params = dict(start_dt=start_dt,
                      end_dt=end_dt,
                      dist=data.get('dist', 250),
                      units=data.get('units', 'km'),
                      min_duration=data.get('min_duration'))
        return {obj.get('title', (obj['lon'], obj['lat'])): await self.intersect(**params, **obj) for obj in objects}

    async def intersect(self, **params):
        return await compute_intersect(self.request.app['redis'], loop=self.request.app.loop, **params)


class LST(RadiusApi):
    async def get(self):
        return web.Response(text=load_html(name='lst'), content_type='text/html')

    async def post(self):
        data = await self.request.post()
        _validate_requires({'start_dt', 'end_dt', 'lst'}, data)
        lst = data['lst']
        objects = read_lst(lst.file.read())

        dts = dt_parser.parse(data['start_dt']), dt_parser.parse(data['end_dt'])
        start_dt, end_dt = min(dts), max(dts)
        if start_dt < datetime(*datetime.today().timetuple()[:3]):
            raise web.HTTPNotFound

        params = dict(start_dt=start_dt,
                      end_dt=end_dt,
                      dist=float(data.get('dist', 250)),
                      units=data.get('units', 'km'),
                      min_duration=int(data.get('min_duration', 60)))
        return {obj.get('title', (obj['lon'], obj['lat'])): await self.intersect(**params, **obj) for obj in objects}


async def index(request):
    return web.Response(text=load_html(), content_type='text/html')


async def get_json(request):
    try:
        data = await request.json()
    except JSONDecodeError:
        raise web.HTTPBadRequest(reason='Wrong JSON format')
    else:
        return data


def _validate_requires(required, iterable, *, is_body=True, msg=None):
    if not required.issubset(iterable):
        raise web.HTTPBadRequest(reason=msg or ('{} parameters must contains required keys `{}`'
                                                .format('Body' if is_body else 'Query', required)))


async def subscribe(request):
    channel_name = request.match_info['channel']
    channel = request.app['channels'][channel_name]
    ws = web.WebSocketResponse()
    await ws.prepare(request)

    channel.add(ws)
    logger.debug('Someone joined to channel "{}".'.format(channel_name))

    try:
        while True:
            msg = await ws.receive_json()
            if msg.get('command') == 'close':
                await ws.close()
    except Exception as exc:
        logger.exception(exc)
    finally:
        channel.remove(ws)

    if ws.closed:
        channel.remove(ws)

    logger.debug('Websocket connection closed in channel "{}"'.format(channel_name))
    return ws


async def compute_intersect(redis, *, start_dt=None, end_dt=None, loop=None, **params):
    min_duration = params.pop('min_duration', None)
    rng = datetime_range(datetime(*start_dt.timetuple()[:4]), datetime(*end_dt.timetuple()[:4]), timedelta(hours=1))
    func = partial(geo_radius, redis=redis, **params)
    result = await asyncio.gather(*(func(dt=dt) for dt in rng), loop=loop)
    return filter(lambda val: len(val) > min_duration if min_duration else val, result)



# -*- coding: utf-8 -*-
from .calculations import *
from .log import LOG_FORMAT
from .redis import *
from .server import ISSPositionerService
from .util import *

__version__ = '1.0.0'

__all__ = (
    calculations.__all__ +
    redis.__all__ +
    util.__all__ +
    ('ISSPositionerService', 'LOG_FORMAT',)
)



# -*- coding: utf-8  -*-
import asyncio
import math
from collections import OrderedDict
from datetime import datetime, date, timedelta

from ephem import readtle

from .util import datetime_range, find_le

__all__ = (
    'AsyncSatellite',
    'Satellite',
    'TLENotFound',
)


class TLENotFound(ValueError):
    pass


class Satellite(object):
    _tle_db = None

    def __init__(self, *, tles=None):
        self.tle_db = tles

    @property
    def tle_db(self):
        return self._tle_db

    @tle_db.setter
    def tle_db(self, values):
        self._tle_db = OrderedDict((self.tle_date(tle), tle) for tle in values or [])

    def get_nearest_tle(self, dt=None, ignore=True):
        return self.tle_db[self.get_nearest_tle_dt(dt, ignore)]

    def get_nearest_tle_dt(self, dt=None, ignore=True):
        tle_dts = tuple(self.tle_db.keys())

        if not isinstance(dt, (datetime, date)):
            return tle_dts[0]

        try:
            tle_dt = find_le(tle_dts, dt)
        except ValueError:
            if ignore:
                return tle_dts[0]
            raise TLENotFound('Not found TLE for %r, the oldest TLE in db for %r' % (dt, tle_dts[0]))
        else:
            return tle_dt

    @staticmethod
    def tle_date(tle):
        fmt = '%Y-%m-%d %H:%M:%S.%f'
        return datetime.strptime('{epoch}.{epoch_microseconds}'.format(**tle), fmt)

    def get(self, dt, *, tle=None):
        if tle is None and isinstance(dt, (datetime, date)):
            tle = self.get_nearest_tle(dt)
        return readtle(tle['tle_line0'], tle['tle_line1'], tle['tle_line2'])

    def get_position(self, dt, *, tle=None):
        iss = self.get(dt, tle=tle)
        iss.compute(dt)
        return math.degrees(iss.sublong), math.degrees(iss.sublat), dt.isoformat(' ')


class AsyncSatellite(Satellite):
    _loop = None

    def __init__(self, *, tles=None, loop=None):
        super().__init__(tles=tles)
        if loop is None:
            loop = asyncio.get_event_loop()

        self._loop = loop

    @property
    def loop(self):
        return self._loop

    async def get_position(self, dt, *, tle=None):
        return super().get_position(dt, tle=tle)

    async def compute(self, start_dt=None, end_dt=None, step=1):
        if not isinstance(start_dt, datetime):
            start_dt = datetime.utcnow()
        if not isinstance(end_dt, datetime):
            end_dt = start_dt + timedelta(days=1)
        fs = (self.get_position(t) for t in datetime_range(start_dt, end_dt, step))
        return await asyncio.gather(*fs, loop=self.loop)



# -*- coding: utf-8 -*-
import asyncio
import os
import re
import ujson
from bisect import bisect_right
from datetime import datetime, timedelta
from functools import wraps

import yaml
from aiohttp import ClientSession

from .log import logger

__all__ = (
    'audit',
    'datetime_range',
    'find_le',
    'geo_member_to_dict',
    'get_tle',
    'load_cfg',
    'load_html',
    'periodic_task',
    'read_lst',
    'PROJECT_DIR',
)

PROJECT_DIR = os.path.join(os.path.dirname(__file__))
LST_EXP = re.compile(r'(?P<lon>[\d+-]+(\.)?\d+?)\s+(?P<lat>[\d+-]+(\.)?\d+?)\s+\'(?P<title>.+?)\'')


def audit(*args, **kwargs):
    def wrapper(func):
        @wraps(func)
        def inner_wrapper(*args, **kwargs):
            dt = datetime.now()
            log.debug('Start execution `%s`', func.__name__)
            result = func(*args, **kwargs)
            log.debug('End execution `%s` (%s)', func.__name__, datetime.now() - dt)
            return result

        @wraps(func)
        async def async_inner_wrapper(*args, **kwargs):
            dt = datetime.now()
            log.debug('Start execution `%s`', func.__name__)
            result = await func(*args, **kwargs)
            log.debug('End execution `%s` (%s)', func.__name__, datetime.now() - dt)
            return result

        log = kwargs.pop('logger', logger)

        return async_inner_wrapper if asyncio.iscoroutinefunction(func) else inner_wrapper

    if args and kwargs:
        raise ValueError("cannot combine positional and keyword args")
    if len(args) == 1:
        return wrapper(args[0])
    elif len(args) != 0:
        raise ValueError("expected 1 argument, got %d", len(args))
    return wrapper


def datetime_range(start, end, step=None):
    if not isinstance(start, datetime) or not isinstance(end, datetime):
        raise ValueError

    if not isinstance(step, timedelta):
        step = timedelta(seconds=step or 1)

    if start == end:
        yield start

    while start < end:
        yield start
        start += step


def find_le(a, x):
    """
    Find rightmost value less than or equal to x
    """
    i = bisect_right(a, x)
    if i:
        return a[i - 1]
    raise ValueError


async def get_tle(*, url=None, loop=None):
    d = datetime.today().date() - timedelta(days=1)
    query = {
        "filters": ujson.dumps({"dt": {"$gte": d.isoformat()}, "norad_cat_id": 25544}),
        "order": "dt",
        "only": "id,source,extra_info"
    }
    async with ClientSession(loop=loop) as session:
        async with session.get(url, params=query) as resp:
            r = await resp.json(loads=ujson.loads)
    return r.get('data')


def geo_member_to_dict(geo_member, *, units=None):
    return dict(dt=geo_member.member,
                dist=geo_member.dist,
                units=units,
                geohash=geo_member.hash,
                coord=geo_member.coord._asdict())


def load_html(*, name='index', path=None):
    if not isinstance(path, str):
        path = os.path.join(PROJECT_DIR, 'html', '{}.{}'.format(name, 'html'))

    if not os.path.exists(path):
        raise FileNotFoundError

    with open(path, 'r') as html:
        return html.read()


def load_cfg(*, filename='dev', path=None):
    if not isinstance(path, str):
        path = os.path.join(PROJECT_DIR, 'config', '{}.{}'.format(filename, 'yaml'))

    if not os.path.exists(path):
        raise FileNotFoundError

    with open(path, 'r') as cfg:
        return yaml.safe_load(cfg)


def periodic_task(delay):
    def wrapper(coroutine):
        @wraps(coroutine)
        async def runner(*args, **kwargs):
            while True:
                await coroutine(*args, **kwargs)
                await asyncio.sleep(delay)

        return runner

    return wrapper


def read_lst(lst, encoding='cp866'):
    if isinstance(lst, bytes):
        lst = lst.decode(encoding)
    objects = []
    for m in LST_EXP.finditer(lst):
        m = m.groupdict()
        m.update(lat=float(m['lat']), lon=float(m['lon']))
        objects.append(m)
    return objects



# -*- coding: utf-8 -*-
import io
import logging
from functools import partial

from tqdm import tqdm

LOG_FORMAT = '[%(levelname)1.1s %(asctime)s %(name)s %(module)s:%(lineno)i] %(message)s'
logger = logging.getLogger('iss_positioner')


__all__ = (
    'logger',
    'LOG_FORMAT',
)

class TqdmToLogger(io.StringIO):
    """
    Output stream for TQDM which will output to logger module instead of the StdOut.
    """
    buf = ''

    def __init__(self, log=logger, level=logging.INFO):
        super().__init__()
        self.logger = log
        self.level = level

    def write(self, buf):
        self.buf = buf.strip('\r\n')

    def flush(self):
        self.logger.log(self.level, self.buf)


tqdm = partial(tqdm, file=TqdmToLogger())



# -*- coding: utf-8 -*-
import traceback
import ujson
from functools import partial

from aiohttp import web

__all__ = (
    'json_response_middleware',
)

json_response = partial(web.json_response, dumps=ujson.dumps)


def json_error(reason, status=500):
    return json_response({'error': True, 'error_msg': reason, 'data': None}, status=status)


def json_success(data):
    return json_response({'error': False, 'error_msg': None, 'data': data})


async def json_response_middleware(app, handler):
    async def middleware_handler(request):
        try:
            data = await handler(request)
            return data if isinstance(data, web.Response) else json_success(data)
        except web.HTTPException as exc:
            return json_error(exc.reason, exc.status_code)
        except:
            return json_error(traceback.format_exc())

    return middleware_handler



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import codecs
import os
import re
import sys

from setuptools import setup, find_packages

if sys.version_info < (3, 5, 0):
    raise RuntimeError("iss-positioner requires Python 3.5.0+")

PROJECT_DIR = os.path.abspath(os.path.dirname(__file__))
VERSION_REGEXP = re.compile(r"^__version__ = [\'\"](.+?)[\'\"]$", re.MULTILINE)


def read(fn):
    with codecs.open(os.path.join(PROJECT_DIR, fn), encoding='utf-8') as f:
        return f.read().strip()


def version():
    try:
        return VERSION_REGEXP.findall(read(os.path.join('iss_positioner', '__init__.py')))[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')


vn = version()
url = 'https://github.com/nkoshell/iss-positioner'

setup(
    name='iss-positioner',
    description='ISS position computing service',
    long_description=read('README.rst'),
    version=vn,
    packages=find_packages(),
    include_package_data=True,
    url=url,
    download_url='{url}/archive/{version}.tar.gz'.format(url=url, version=vn),
    license='MIT',
    author='nkoshell',
    author_email='nikita.koshelev@gmail.com',
    install_requires=read('requirements.in').splitlines(),
)
